#!/usr/bin/env python3
# Copyright 2024 Alexander Larsson
# SPDX-License-Identifier: LGPL-2.1-or-later
#
# Integration tests for bwrap sandbox setup. Methods decorated with
# @in_sandbox(...) define assertions that run inside the sandbox;
# the @sandbox_test_class decorator auto-generates the host-side
# test_* methods that launch bwrap and run them.

import ctypes
import ctypes.util
import importlib.util
import os
import stat
import subprocess
import sys
import tempfile
import unittest

_spec = importlib.util.spec_from_file_location(
    'test_helper',
    os.path.join(os.path.dirname(os.path.abspath(__file__)), 'test-helper.py'),
)
_helper = importlib.util.module_from_spec(_spec)
_spec.loader.exec_module(_helper)

BWRAP = _helper.BWRAP
can_run_bwrap = _helper.can_run_bwrap
in_sandbox = _helper.in_sandbox
list_mounts = _helper.list_mounts
DataFd = _helper.DataFd
OpenFd = _helper.OpenFd
run_bwrap = _helper.run_bwrap
run_tap_tests = _helper.run_tap_tests
sandbox_test_class = _helper.sandbox_test_class


class SourceTree:
    """Creates a temporary tree with various file types for testing."""

    def __init__(self, tmp_dir):
        self.root = os.path.join(tmp_dir, 'src')
        os.makedirs(self.root)

        self.file = os.path.join(self.root, 'regular_file')
        with open(self.file, 'w') as f:
            f.write('hello')

        self.other_file = os.path.join(self.root, 'other_file')
        with open(self.other_file, 'w') as f:
            f.write('other')

        self.dir = os.path.join(self.root, 'subdir')
        os.makedirs(self.dir)
        with open(os.path.join(self.dir, 'child'), 'w') as f:
            f.write('child content')

        self.dir2 = os.path.join(self.root, 'subdir2')
        os.makedirs(self.dir2)
        with open(os.path.join(self.dir2, 'child2'), 'w') as f:
            f.write('child2 content')

        self.symlink_to_file = os.path.join(self.root, 'link_to_file')
        os.symlink(self.file, self.symlink_to_file)

        self.symlink_to_dir = os.path.join(self.root, 'link_to_dir')
        os.symlink(self.dir, self.symlink_to_dir)

        self.symlink_relative = os.path.join(self.root, 'link_relative')
        os.symlink('regular_file', self.symlink_relative)


@sandbox_test_class
@unittest.skipUnless(can_run_bwrap(), 'bwrap not available or not functional')
class TestSandbox(unittest.TestCase):
    """Integration tests for bwrap sandbox setup options."""

    def setUp(self):
        self._tmp = tempfile.TemporaryDirectory()
        self.tmp = self._tmp.name
        self.src = SourceTree(self.tmp)

    def tearDown(self):
        self._tmp.cleanup()

    def assertIsRegFile(self, path):
        self.assertTrue(stat.S_ISREG(os.lstat(path).st_mode),
                        f'{path} is not a regular file')

    def assertIsChrDev(self, path):
        self.assertTrue(stat.S_ISCHR(os.lstat(path).st_mode),
                        f'{path} is not a character device')

    def assertIsDir(self, path):
        self.assertTrue(stat.S_ISDIR(os.lstat(path).st_mode),
                        f'{path} is not a directory')

    def assertIsSymlink(self, path):
        self.assertTrue(stat.S_ISLNK(os.lstat(path).st_mode),
                        f'{path} is not a symlink')

    def assertSymlinkContent(self, path, expected):
        self.assertIsSymlink(path)
        self.assertEqual(os.readlink(path), expected,
                         f'{path} points to {os.readlink(path)!r}, expected {expected!r}')

    def assertInStderr(self, needle, result):
        self.assertIn(needle, result.stderr,
                      f'expected {needle!r} in stderr, got: {result.stderr!r}')

    def assertFileContent(self, path, expected):
        with open(path) as f:
            self.assertEqual(f.read(), expected)

    _libc = ctypes.CDLL(ctypes.util.find_library('c'), use_errno=True)
    TMPFS_MAGIC = 0x01021994

    def _statfs_type(self, path):
        class statfs_t(ctypes.Structure):
            _fields_ = [('f_type', ctypes.c_long), ('f_bsize', ctypes.c_long)]
        buf = statfs_t()
        rc = self._libc.statfs(path.encode(), ctypes.byref(buf))
        self.assertEqual(rc, 0, f'statfs({path}) failed')
        return buf.f_type

    def assertIsTmpfs(self, path):
        fs_type = self._statfs_type(path)
        self.assertEqual(fs_type, self.TMPFS_MAGIC,
                         f'{path} is not tmpfs (f_type=0x{fs_type:x})')

    def assertIsMountpoint(self, path):
        self.assertIn(path, list_mounts(),
                      f'{path} is not a mountpoint')

    def assertBwrapFailed(self, result):
        self.assertNotEqual(result.returncode, 0,
                            f'expected bwrap to fail, stderr: {result.stderr!r}')

    def assertMountFlags(self, path, expected, unexpected):
        flags = os.statvfs(path).f_flag
        for f in expected:
            self.assertTrue(flags & f, f'{path}: expected flag {f:#x} not set')
        for f in unexpected:
            self.assertFalse(flags & f, f'{path}: unexpected flag {f:#x} set')

    # ------ Source is a regular file ------

    @in_sandbox('--ro-bind', lambda self: self.src.file, '/tmp/f')
    def test_ro_bind_file(self):
        self.assertIsRegFile('/tmp/f')
        self.assertFileContent('/tmp/f', 'hello')
        self.assertIsMountpoint('/tmp/f')
        self.assertMountFlags('/tmp/f',
                              [os.ST_RDONLY, os.ST_NOSUID, os.ST_NODEV], [])

    @in_sandbox('--bind', lambda self: self.src.file, '/tmp/f')
    def test_bind_file(self):
        self.assertIsRegFile('/tmp/f')
        self.assertFileContent('/tmp/f', 'hello')
        self.assertIsMountpoint('/tmp/f')
        self.assertMountFlags('/tmp/f',
                              [os.ST_NOSUID, os.ST_NODEV], [os.ST_RDONLY])

    @in_sandbox('--dev-bind', '/dev/null', '/tmp/devnull')
    def test_dev_bind_file(self):
        self.assertIsChrDev('/tmp/devnull')
        self.assertIsMountpoint('/tmp/devnull')
        self.assertMountFlags('/tmp/devnull',
                              [os.ST_NOSUID], [os.ST_RDONLY, os.ST_NODEV])

    # ------ Source is a directory ------

    @in_sandbox('--ro-bind', lambda self: self.src.dir, '/tmp/d')
    def test_ro_bind_dir(self):
        self.assertIsDir('/tmp/d')
        self.assertFileContent('/tmp/d/child', 'child content')
        self.assertIsMountpoint('/tmp/d')
        self.assertMountFlags('/tmp/d',
                              [os.ST_RDONLY, os.ST_NOSUID, os.ST_NODEV], [])

    @in_sandbox('--bind', lambda self: self.src.dir, '/tmp/d')
    def test_bind_dir(self):
        self.assertIsDir('/tmp/d')
        self.assertFileContent('/tmp/d/child', 'child content')
        self.assertIsMountpoint('/tmp/d')
        self.assertMountFlags('/tmp/d',
                              [os.ST_NOSUID, os.ST_NODEV], [os.ST_RDONLY])

    @in_sandbox('--dev-bind', '/dev', '/tmp/d')
    def test_dev_bind_dir(self):
        self.assertIsDir('/tmp/d')
        self.assertIsMountpoint('/tmp/d')
        self.assertMountFlags('/tmp/d',
                              [os.ST_NOSUID], [os.ST_RDONLY, os.ST_NODEV])

    # ------ Source is a symlink (should resolve to target) ------

    @in_sandbox('--ro-bind', lambda self: self.src.symlink_to_file, '/tmp/f')
    def test_ro_bind_symlink_to_file(self):
        self.assertIsRegFile('/tmp/f')
        self.assertFileContent('/tmp/f', 'hello')
        self.assertMountFlags('/tmp/f',
                              [os.ST_RDONLY, os.ST_NOSUID, os.ST_NODEV], [])

    @in_sandbox('--ro-bind', lambda self: self.src.symlink_to_dir, '/tmp/d')
    def test_ro_bind_symlink_to_dir(self):
        self.assertIsDir('/tmp/d')
        self.assertFileContent('/tmp/d/child', 'child content')
        self.assertMountFlags('/tmp/d',
                              [os.ST_RDONLY, os.ST_NOSUID, os.ST_NODEV], [])

    @in_sandbox('--ro-bind', lambda self: self.src.symlink_relative, '/tmp/f')
    def test_ro_bind_relative_symlink(self):
        self.assertIsRegFile('/tmp/f')
        self.assertFileContent('/tmp/f', 'hello')
        self.assertMountFlags('/tmp/f',
                              [os.ST_RDONLY, os.ST_NOSUID, os.ST_NODEV], [])

    # ------ Destination parent creation ------

    @in_sandbox('--ro-bind', lambda self: self.src.file,
                '/tmp/deep/nested/path/file')
    def test_bind_creates_parent_dirs(self):
        self.assertIsDir('/tmp/deep/nested/path')
        self.assertFileContent('/tmp/deep/nested/path/file', 'hello')

    @in_sandbox('--ro-bind', lambda self: self.src.dir,
                '/tmp/deep/nested/dir')
    def test_bind_dir_creates_parent_dirs(self):
        self.assertIsDir('/tmp/deep/nested/dir')
        self.assertFileContent('/tmp/deep/nested/dir/child', 'child content')

    # ------ Destination already exists ------

    @in_sandbox('--dir', '/tmp/existing',
                '--ro-bind', lambda self: self.src.dir, '/tmp/existing')
    def test_bind_over_existing_dir(self):
        self.assertFileContent('/tmp/existing/child', 'child content')

    @in_sandbox('--ro-bind', lambda self: self.src.other_file, '/tmp/f',
                '--ro-bind', lambda self: self.src.file, '/tmp/f')
    def test_bind_over_existing_file(self):
        self.assertFileContent('/tmp/f', 'hello')

    # ------ Destination is a symlink (should fail) ------

    def test_bind_to_symlink_dest_fails(self):
        result = run_bwrap('--dir', '/tmp/dir',
                           '--symlink', 'dir', '/tmp/link',
                           '--ro-bind', self.src.dir, '/tmp/link',
                           'true')
        self.assertBwrapFailed(result)
        self.assertInStderr(b"Can't mount on symlink destination", result)

    def test_bind_to_symlink_dest_fails_file(self):
        result = run_bwrap('--symlink', 'dir', '/tmp/link',
                           '--ro-bind', self.src.file, '/tmp/link',
                           'true')
        self.assertBwrapFailed(result)
        self.assertInStderr(b"Can't mount on symlink destination", result)

    # ------ bind-fd / ro-bind-fd ------

    @in_sandbox('--ro-bind-fd', lambda self: OpenFd(self.src.file), '/tmp/f')
    def test_ro_bind_fd_file(self):
        self.assertIsRegFile('/tmp/f')
        self.assertFileContent('/tmp/f', 'hello')
        self.assertMountFlags('/tmp/f',
                              [os.ST_RDONLY, os.ST_NOSUID, os.ST_NODEV], [])

    @in_sandbox('--ro-bind-fd', lambda self: OpenFd(self.src.dir), '/tmp/d')
    def test_ro_bind_fd_dir(self):
        self.assertIsDir('/tmp/d')
        self.assertFileContent('/tmp/d/child', 'child content')
        self.assertMountFlags('/tmp/d',
                              [os.ST_RDONLY, os.ST_NOSUID, os.ST_NODEV], [])

    @in_sandbox('--bind-fd', lambda self: OpenFd(self.src.file), '/tmp/f')
    def test_bind_fd_file(self):
        self.assertIsRegFile('/tmp/f')
        self.assertFileContent('/tmp/f', 'hello')
        self.assertMountFlags('/tmp/f',
                              [os.ST_NOSUID, os.ST_NODEV], [os.ST_RDONLY])

    @in_sandbox('--bind-fd', lambda self: OpenFd(self.src.dir), '/tmp/d')
    def test_bind_fd_dir(self):
        self.assertIsDir('/tmp/d')
        self.assertFileContent('/tmp/d/child', 'child content')
        self.assertMountFlags('/tmp/d',
                              [os.ST_NOSUID, os.ST_NODEV], [os.ST_RDONLY])

    # ------ Multiple binds ------

    @in_sandbox('--ro-bind', lambda self: self.src.file, '/tmp/a',
                '--ro-bind', lambda self: self.src.dir, '/tmp/b')
    def test_multiple_binds(self):
        self.assertFileContent('/tmp/a', 'hello')
        self.assertIsDir('/tmp/b')
        self.assertFileContent('/tmp/b/child', 'child content')

    @in_sandbox('--ro-bind', lambda self: self.src.file, '/tmp/parent/a',
                '--ro-bind', lambda self: self.src.dir, '/tmp/parent/b')
    def test_bind_file_then_dir_same_parent(self):
        self.assertFileContent('/tmp/parent/a', 'hello')
        self.assertIsDir('/tmp/parent/b')

    # ------ Recursive mount flags ------

    @in_sandbox()
    def test_ro_bind_root_is_recursive(self):
        """The base --ro-bind / / should apply read-only to sub-mounts."""
        skip_prefixes = ('/dev', '/proc', '/sys', '/tmp')
        for mountpoint in list_mounts():
            if any(mountpoint == p or mountpoint.startswith(p + '/')
                   for p in skip_prefixes):
                continue
            flags = os.statvfs(mountpoint).f_flag
            self.assertTrue(
                flags & os.ST_RDONLY,
                f'{mountpoint} is not read-only',
            )

    # ------ tmpfs ------

    @in_sandbox('--tmpfs', '/tmp/t')
    def test_tmpfs(self):
        self.assertIsDir('/tmp/t')
        self.assertIsMountpoint('/tmp/t')
        self.assertIsTmpfs('/tmp/t')
        self.assertMountFlags('/tmp/t',
                              [os.ST_NOSUID, os.ST_NODEV], [os.ST_RDONLY])
        with open('/tmp/t/testfile', 'w') as f:
            f.write('writable')
        self.assertFileContent('/tmp/t/testfile', 'writable')

    @in_sandbox('--perms', '0700', '--tmpfs', '/tmp/t')
    def test_tmpfs_perms(self):
        self.assertIsMountpoint('/tmp/t')
        self.assertIsTmpfs('/tmp/t')
        mode = os.lstat('/tmp/t').st_mode & 0o7777
        self.assertEqual(mode, 0o700)

    @in_sandbox('--size', str(64 * 1024), '--tmpfs', '/tmp/t')
    def test_tmpfs_size(self):
        self.assertIsMountpoint('/tmp/t')
        self.assertIsTmpfs('/tmp/t')
        st = os.statvfs('/tmp/t')
        total = st.f_blocks * st.f_frsize
        self.assertLessEqual(total, 64 * 1024 + 4096)

    # ------ proc ------

    @in_sandbox()
    def test_proc(self):
        self.assertIsSymlink('/proc/self')
        self.assertIsDir('/proc/self/')
        self.assertIsRegFile('/proc/self/status')
        for subdir in ['sys', 'sysrq-trigger', 'irq', 'bus']:
            path = f'/proc/{subdir}'
            if os.path.exists(path):
                self.assertFalse(os.access(path, os.W_OK),
                                 f'/proc/{subdir} should not be writable')

    # ------ dev ------

    @in_sandbox()
    def test_dev_nodes(self):
        for node in ['null', 'zero', 'full', 'random', 'urandom', 'tty']:
            self.assertIsChrDev(f'/dev/{node}')
        for link in ['stdin', 'stdout', 'stderr', 'fd', 'ptmx']:
            self.assertIsSymlink(f'/dev/{link}')
        self.assertIsDir('/dev/pts')
        self.assertIsDir('/dev/shm')

    # ------ dir ------

    @in_sandbox('--dir', '/tmp/d')
    def test_dir(self):
        self.assertIsDir('/tmp/d')

    @in_sandbox('--perms', '0700', '--dir', '/tmp/d')
    def test_dir_perms(self):
        self.assertIsDir('/tmp/d')
        self.assertEqual(os.lstat('/tmp/d').st_mode & 0o7777, 0o700)

    @in_sandbox('--dir', '/tmp/a/b/c')
    def test_dir_creates_parents(self):
        self.assertIsDir('/tmp/a')
        self.assertIsDir('/tmp/a/b')
        self.assertIsDir('/tmp/a/b/c')

    # ------ file ------

    @in_sandbox('--file', DataFd(b'file content'), '/tmp/f')
    def test_file(self):
        self.assertIsRegFile('/tmp/f')
        self.assertFileContent('/tmp/f', 'file content')

    @in_sandbox('--perms', '0600', '--file', DataFd(b'x'), '/tmp/f')
    def test_file_perms(self):
        self.assertIsRegFile('/tmp/f')
        self.assertEqual(os.lstat('/tmp/f').st_mode & 0o7777, 0o600)

    # ------ bind-data / ro-bind-data ------

    @in_sandbox('--bind-data', DataFd(b'bind data'), '/tmp/f')
    def test_bind_data(self):
        self.assertIsRegFile('/tmp/f')
        self.assertFileContent('/tmp/f', 'bind data')
        self.assertIsMountpoint('/tmp/f')
        self.assertMountFlags('/tmp/f',
                              [os.ST_NOSUID, os.ST_NODEV], [os.ST_RDONLY])

    @in_sandbox('--ro-bind-data', DataFd(b'ro data'), '/tmp/f')
    def test_ro_bind_data(self):
        self.assertIsRegFile('/tmp/f')
        self.assertFileContent('/tmp/f', 'ro data')
        self.assertIsMountpoint('/tmp/f')
        self.assertMountFlags('/tmp/f',
                              [os.ST_RDONLY, os.ST_NOSUID, os.ST_NODEV], [])

    # ------ symlink ------

    @in_sandbox('--symlink', '/usr', '/tmp/link')
    def test_symlink(self):
        self.assertSymlinkContent('/tmp/link', '/usr')

    @in_sandbox('--symlink', '/usr', '/tmp/link',
                '--symlink', '/usr', '/tmp/link')
    def test_symlink_idempotent(self):
        self.assertSymlinkContent('/tmp/link', '/usr')

    def test_symlink_conflict(self):
        result = run_bwrap('--symlink', '/usr', '/tmp/link',
                           '--symlink', '/etc', '/tmp/link',
                           'true')
        self.assertBwrapFailed(result)
        self.assertInStderr(b'existing destination is /usr', result)

    def test_symlink_over_file(self):
        result = run_bwrap('--file', DataFd(b''), '/tmp/f',
                           '--symlink', '/usr', '/tmp/f',
                           'true')
        self.assertBwrapFailed(result)
        self.assertInStderr(b'not a symlink', result)

    # ------ chmod ------

    @in_sandbox('--dir', '/tmp/d', '--chmod', '0700', '/tmp/d')
    def test_chmod(self):
        self.assertEqual(os.lstat('/tmp/d').st_mode & 0o7777, 0o700)

    @in_sandbox('--dir', '/tmp/d', '--chmod', '0755', '/tmp/d')
    def test_chmod_755(self):
        self.assertEqual(os.lstat('/tmp/d').st_mode & 0o7777, 0o755)

    @in_sandbox('--file', DataFd(b'x'), '/tmp/f', '--chmod', '0600', '/tmp/f')
    def test_chmod_file(self):
        self.assertEqual(os.lstat('/tmp/f').st_mode & 0o7777, 0o600)

    def test_chmod_nonexistent(self):
        result = run_bwrap('--chmod', '0700', '/tmp/noexist', 'true')
        self.assertBwrapFailed(result)
        self.assertInStderr(b"No such file or directory", result)

    # ------ remount-ro ------

    @in_sandbox('--tmpfs', '/tmp/rw', '--remount-ro', '/tmp/rw')
    def test_remount_ro(self):
        self.assertMountFlags('/tmp/rw', [os.ST_RDONLY], [])

    # ------ mqueue ------

    def test_mqueue(self):
        result = run_bwrap('--mqueue', '/tmp/mq', 'true')
        if result.returncode != 0 and b'Operation not permitted' in result.stderr:
            self.skipTest('mqueue not available in user namespace')
        self.assertEqual(result.returncode, 0)

    # ------ hostname ------

    def test_hostname(self):
        result = run_bwrap('--unshare-uts', '--hostname', 'testhost',
                           'hostname')
        self.assertEqual(result.returncode, 0)
        self.assertEqual(result.stdout.strip(), b'testhost')

    # ------ bind-try / ro-bind-try (nonexistent source skipped) ------

    @in_sandbox('--ro-bind-try', '/nonexistent/path', '/tmp/f')
    def test_ro_bind_try_nonexistent(self):
        self.assertFalse(os.path.exists('/tmp/f'))

    @in_sandbox('--bind-try', '/nonexistent/path', '/tmp/f')
    def test_bind_try_nonexistent(self):
        self.assertFalse(os.path.exists('/tmp/f'))

    @in_sandbox('--dev-bind-try', '/nonexistent/path', '/tmp/f')
    def test_dev_bind_try_nonexistent(self):
        self.assertFalse(os.path.exists('/tmp/f'))

    @in_sandbox('--ro-bind-try', lambda self: self.src.file, '/tmp/f')
    def test_ro_bind_try_existing(self):
        self.assertFileContent('/tmp/f', 'hello')

    # ------ overlay ------

    def test_tmp_overlay(self):
        result = run_bwrap('--overlay-src', self.src.dir,
                           '--tmp-overlay', '/tmp/ov',
                           'cat', '/tmp/ov/child')
        if result.returncode != 0 and b'overlay' in result.stderr.lower():
            self.skipTest('overlayfs not available')
        self.assertEqual(result.returncode, 0)
        self.assertEqual(result.stdout, b'child content')

    def test_ro_overlay(self):
        result = run_bwrap('--overlay-src', self.src.dir,
                           '--overlay-src', self.src.dir2,
                           '--ro-overlay', '/tmp/ov',
                           'cat', '/tmp/ov/child')
        if result.returncode != 0 and b'overlay' in result.stderr.lower():
            self.skipTest('overlayfs not available')
        self.assertEqual(result.returncode, 0)
        self.assertEqual(result.stdout, b'child content')

    # ------ Edge cases: source path with symlinks ------

    def test_bind_source_absolute_symlink_in_path(self):
        """Source path contains a symlink with absolute target."""
        link = os.path.join(self.tmp, 'abs_link')
        os.symlink(self.src.root, link)
        result = run_bwrap('--ro-bind', link + '/regular_file', '/tmp/f',
                           'cat', '/tmp/f')
        self.assertEqual(result.returncode, 0)
        self.assertEqual(result.stdout, b'hello')

    def test_bind_source_dotdot_in_path(self):
        """Source path with .. components."""
        path = self.src.dir + '/../regular_file'
        result = run_bwrap('--ro-bind', path, '/tmp/f',
                           'cat', '/tmp/f')
        self.assertEqual(result.returncode, 0)
        self.assertEqual(result.stdout, b'hello')

    def test_bind_source_symlink_chain(self):
        """Source is a chain of symlinks."""
        link1 = os.path.join(self.tmp, 'link1')
        link2 = os.path.join(self.tmp, 'link2')
        os.symlink(self.src.file, link1)
        os.symlink(link1, link2)
        result = run_bwrap('--ro-bind', link2, '/tmp/f',
                           'cat', '/tmp/f')
        self.assertEqual(result.returncode, 0)
        self.assertEqual(result.stdout, b'hello')

    # ------ Edge cases: dest parent issues ------

    def test_bind_dest_parent_is_file(self):
        """Bind where a parent component of dest is an existing file."""
        result = run_bwrap('--file', DataFd(b''), '/tmp/notadir',
                           '--ro-bind', self.src.file, '/tmp/notadir/child',
                           'true')
        self.assertBwrapFailed(result)
        self.assertInStderr(b"Not a directory", result)

    @in_sandbox('--dir', '/tmp/target',
                '--symlink', 'target', '/tmp/link',
                '--ro-bind', lambda self: self.src.file, '/tmp/link/child')
    def test_bind_dest_parent_is_symlink(self):
        """Bind where a parent component of dest is a symlink to a dir."""
        self.assertFileContent('/tmp/link/child', 'hello')

    def test_dir_dest_parent_is_file(self):
        """--dir where a parent component is an existing file."""
        result = run_bwrap('--file', DataFd(b''), '/tmp/notadir',
                           '--dir', '/tmp/notadir/sub',
                           'true')
        self.assertBwrapFailed(result)
        self.assertInStderr(b"Not a directory", result)

    # ------ Edge cases: dest type mismatches ------

    def test_bind_dir_over_existing_file(self):
        """Bind a directory source over an existing file dest."""
        result = run_bwrap('--file', DataFd(b''), '/tmp/f',
                           '--ro-bind', self.src.dir, '/tmp/f',
                           'true')
        self.assertBwrapFailed(result)
        self.assertInStderr(b'not a directory', result)

    def test_bind_file_over_existing_dir(self):
        """Bind a file source over an existing directory dest."""
        result = run_bwrap('--dir', '/tmp/d',
                           '--ro-bind', self.src.file, '/tmp/d',
                           'true')
        self.assertBwrapFailed(result)
        self.assertInStderr(b"not a file", result)

    # ------ Edge cases: nonexistent source ------

    def test_bind_nonexistent_source(self):
        result = run_bwrap('--ro-bind', '/nonexistent/path', '/tmp/f',
                           'true')
        self.assertBwrapFailed(result)
        self.assertInStderr(b"No such file or directory", result)

    # ------ Edge cases: multiple overlapping mounts ------

    @in_sandbox('--tmpfs', '/tmp/a',
                '--dir', '/tmp/a/sub',
                '--ro-bind', lambda self: self.src.file, '/tmp/a/sub/f')
    def test_mount_chain(self):
        """tmpfs, then dir inside it, then bind inside that."""
        self.assertIsDir('/tmp/a/sub')
        self.assertFileContent('/tmp/a/sub/f', 'hello')

    # ------ Edge cases: bind-fd dest is symlink ------

    def test_bind_fd_to_symlink_dest_fails(self):
        fd = os.open(self.src.dir, os.O_PATH)
        result = run_bwrap('--dir', '/tmp/dir',
                           '--symlink', 'dir', '/tmp/link',
                           '--bind-fd', str(fd), '/tmp/link',
                           'true', pass_fds=(fd,))
        os.close(fd)
        self.assertBwrapFailed(result)
        self.assertInStderr(b"Can't mount on symlink destination", result)

    # ------ Edge cases: --file creates parents ------

    @in_sandbox('--file', DataFd(b'deep'), '/tmp/a/b/c')
    def test_file_creates_parents(self):
        self.assertIsDir('/tmp/a/b')
        self.assertFileContent('/tmp/a/b/c', 'deep')

    # ------ Edge cases: --symlink creates parents ------

    @in_sandbox('--symlink', '/usr', '/tmp/a/b/link')
    def test_symlink_creates_parents(self):
        self.assertIsDir('/tmp/a/b')
        self.assertSymlinkContent('/tmp/a/b/link', '/usr')

    # ------ Edge cases: --dir over existing dir is fine ------

    @in_sandbox('--dir', '/tmp/d', '--dir', '/tmp/d')
    def test_dir_idempotent(self):
        self.assertIsDir('/tmp/d')

    # ------ Edge cases: bind-data creates parents ------

    @in_sandbox('--ro-bind-data', DataFd(b'nested'), '/tmp/a/b/f')
    def test_bind_data_creates_parents(self):
        self.assertIsDir('/tmp/a/b')
        self.assertFileContent('/tmp/a/b/f', 'nested')

    # ------ Symlink escape via /proc/self/fd ------

    def _test_proc_symlink_escape(self, extra_bwrap_args):
        """Verifies that a symlink pointing through /proc/self/fd
        to an fd outside the sandbox cannot be used to escape the root."""
        escape_target = os.path.join(self.tmp, 'escape_target')
        os.makedirs(escape_target)

        escape_fd = os.open(escape_target, os.O_PATH | os.O_DIRECTORY)
        try:
            result = run_bwrap(
                *extra_bwrap_args,
                '--dir', '/tmp/mnt',
                '--symlink', f'/proc/self/fd/{escape_fd}', '/tmp/mnt/symlink',
                '--dir', '/tmp/mnt/symlink/created',
                'true',
                pass_fds=(escape_fd,))

            self.assertBwrapFailed(result)
            self.assertFalse(os.path.exists(
                os.path.join(escape_target, 'created')),
                'directory was created outside sandbox via /proc/self/fd escape')
        finally:
            os.close(escape_fd)

    def test_proc_symlink_escape_blocked(self):
        self._test_proc_symlink_escape([])

    def test_proc_symlink_escape_blocked_fallback(self):
        self._test_proc_symlink_escape(['--debug-opt=force-openat-fallback'])


if __name__ == '__main__':
    run_tap_tests(sys.modules[__name__])



#!/usr/bin/env python3

import os, select, subprocess, sys, json

pipe_info = os.pipe()
userns_block = os.pipe()

pid = os.fork()

if pid != 0:
    os.close(pipe_info[1])
    os.close(userns_block[0])

    select.select([pipe_info[0]], [], [])

    data = json.load(os.fdopen(pipe_info[0]))
    child_pid = str(data['child-pid'])

    subprocess.call(["newuidmap", child_pid, "0", str(os.getuid()), "1"])
    subprocess.call(["newgidmap", child_pid, "0", str(os.getgid()), "1"])

    os.write(userns_block[1], b'1')
else:
    os.close(pipe_info[0])
    os.close(userns_block[1])

    os.set_inheritable(pipe_info[1], True)
    os.set_inheritable(userns_block[0], True)

    args = ["bwrap",
            "bwrap",
            "--unshare-all",
            "--unshare-user",
            "--userns-block-fd", "%i" % userns_block[0],
            "--info-fd", "%i" % pipe_info[1],
            "--bind", "/", "/",
            "cat", "/proc/self/uid_map"]

    os.execlp(*args)



#!/usr/bin/env python3
# Copyright 2021 Simon McVittie
# SPDX-License-Identifier: LGPL-2.1-or-later

import errno
import logging
import os
import subprocess
import sys
import tempfile
import termios
import unittest

try:
    import seccomp
except ImportError:
    print('1..0 # SKIP cannot import seccomp Python module')
    sys.exit(0)


# This is the @default set from systemd as of 2021-10-11
DEFAULT_SET = set('''
brk
cacheflush
clock_getres
clock_getres_time64
clock_gettime
clock_gettime64
clock_nanosleep
clock_nanosleep_time64
execve
exit
exit_group
futex
futex_time64
get_robust_list
get_thread_area
getegid
getegid32
geteuid
geteuid32
getgid
getgid32
getgroups
getgroups32
getpgid
getpgrp
getpid
getppid
getrandom
getresgid
getresgid32
getresuid
getresuid32
getrlimit
getsid
gettid
gettimeofday
getuid
getuid32
membarrier
mmap
mmap2
munmap
nanosleep
pause
prlimit64
restart_syscall
rseq
rt_sigreturn
sched_getaffinity
sched_yield
set_robust_list
set_thread_area
set_tid_address
set_tls
sigreturn
time
ugetrlimit
'''.split())

# This is the @basic-io set from systemd
BASIC_IO_SET = set('''
_llseek
close
close_range
dup
dup2
dup3
lseek
pread64
preadv
preadv2
pwrite64
pwritev
pwritev2
read
readv
write
writev
'''.split())

# This is the @filesystem-io set from systemd
FILESYSTEM_SET = set('''
access
chdir
chmod
close
creat
faccessat
faccessat2
fallocate
fchdir
fchmod
fchmodat
fcntl
fcntl64
fgetxattr
flistxattr
fremovexattr
fsetxattr
fstat
fstat64
fstatat64
fstatfs
fstatfs64
ftruncate
ftruncate64
futimesat
getcwd
getdents
getdents64
getxattr
inotify_add_watch
inotify_init
inotify_init1
inotify_rm_watch
lgetxattr
link
linkat
listxattr
llistxattr
lremovexattr
lsetxattr
lstat
lstat64
mkdir
mkdirat
mknod
mknodat
newfstatat
oldfstat
oldlstat
oldstat
open
openat
openat2
readlink
readlinkat
removexattr
rename
renameat
renameat2
rmdir
setxattr
stat
stat64
statfs
statfs64
statx
symlink
symlinkat
truncate
truncate64
unlink
unlinkat
utime
utimensat
utimensat_time64
utimes
'''.split())

# Miscellaneous syscalls used during process startup, at least on x86_64
ALLOWED = DEFAULT_SET | BASIC_IO_SET | FILESYSTEM_SET | set('''
arch_prctl
ioctl
madvise
mprotect
mremap
prctl
readdir
umask
'''.split())

# Syscalls we will try to use, expecting them to be either allowed or
# blocked by our allow and/or deny lists
TRY_SYSCALLS = [
    'chmod',
    'chroot',
    'clone3',
    'ioctl TIOCNOTTY',
    'ioctl TIOCSTI CVE-2019-10063',
    'ioctl TIOCSTI',
    'listen',
    'prctl',
]


class Test(unittest.TestCase):
    def setUp(self) -> None:
        here = os.path.dirname(os.path.abspath(__file__))

        if 'G_TEST_SRCDIR' in os.environ:
            self.test_srcdir = os.getenv('G_TEST_SRCDIR') + '/tests'
        else:
            self.test_srcdir = here

        if 'G_TEST_BUILDDIR' in os.environ:
            self.test_builddir = os.getenv('G_TEST_BUILDDIR') + '/tests'
        else:
            self.test_builddir = here

        self.bwrap = os.getenv('BWRAP', 'bwrap')
        self.try_syscall = os.path.join(self.test_builddir, 'try-syscall')

        completed = subprocess.run(
            [
                self.bwrap,
                '--ro-bind', '/', '/',
                'true',
            ],
            stdin=subprocess.DEVNULL,
            stdout=subprocess.DEVNULL,
            stderr=2,
        )

        if completed.returncode != 0:
            raise unittest.SkipTest(
                'cannot run bwrap (does it need to be setuid?)'
            )

    def tearDown(self) -> None:
        pass

    def test_no_seccomp(self) -> None:
        for syscall in TRY_SYSCALLS:
            print('# {} without seccomp'.format(syscall))
            completed = subprocess.run(
                [
                    self.bwrap,
                    '--ro-bind', '/', '/',
                    self.try_syscall, syscall,
                ],
                stdin=subprocess.DEVNULL,
                stdout=subprocess.DEVNULL,
                stderr=2,
            )

            if (
                syscall == 'ioctl TIOCSTI CVE-2019-10063'
                and completed.returncode == errno.ENOENT
            ):
                print('# Cannot test 64-bit syscall parameter on 32-bit')
                continue

            if syscall == 'clone3':
                # If the kernel supports it, we didn't block it so
                # it fails with EFAULT. If the kernel doesn't support it,
                # it'll fail with ENOSYS instead.
                self.assertIn(
                    completed.returncode,
                    (errno.ENOSYS, errno.EFAULT),
                )
            elif syscall.startswith('ioctl') or syscall == 'listen':
                self.assertEqual(completed.returncode, errno.EBADF)
            else:
                self.assertEqual(completed.returncode, errno.EFAULT)

    def test_seccomp_allowlist(self) -> None:
        with tempfile.TemporaryFile() as allowlist_temp:
            allowlist = seccomp.SyscallFilter(seccomp.ERRNO(errno.ENOSYS))

            if os.uname().machine == 'x86_64':
                # Allow Python and try-syscall to be different word sizes
                allowlist.add_arch(seccomp.Arch.X86)

            for syscall in ALLOWED:
                try:
                    allowlist.add_rule(seccomp.ALLOW, syscall)
                except Exception as e:
                    print('# Cannot add {} to allowlist: {!r}'.format(syscall, e))

            allowlist.export_bpf(allowlist_temp)

            for syscall in TRY_SYSCALLS:
                print('# allowlist vs. {}'.format(syscall))
                allowlist_temp.seek(0, os.SEEK_SET)

                completed = subprocess.run(
                    [
                        self.bwrap,
                        '--ro-bind', '/', '/',
                        '--seccomp', str(allowlist_temp.fileno()),
                        self.try_syscall, syscall,
                    ],
                    pass_fds=(allowlist_temp.fileno(),),
                    stdin=subprocess.DEVNULL,
                    stdout=subprocess.DEVNULL,
                    stderr=2,
                )

                if (
                    syscall == 'ioctl TIOCSTI CVE-2019-10063'
                    and completed.returncode == errno.ENOENT
                ):
                    print('# Cannot test 64-bit syscall parameter on 32-bit')
                    continue

                if syscall.startswith('ioctl'):
                    # We allow this, so it is executed (and in this simple
                    # example, immediately fails)
                    self.assertEqual(completed.returncode, errno.EBADF)
                elif syscall in ('chroot', 'listen', 'clone3'):
                    # We don't allow these, so they fail with ENOSYS.
                    # clone3 might also be failing with ENOSYS because
                    # the kernel genuinely doesn't support it.
                    self.assertEqual(completed.returncode, errno.ENOSYS)
                else:
                    # We allow this, so it is executed (and in this simple
                    # example, immediately fails)
                    self.assertEqual(completed.returncode, errno.EFAULT)

    def test_seccomp_denylist(self) -> None:
        with tempfile.TemporaryFile() as denylist_temp:
            denylist = seccomp.SyscallFilter(seccomp.ALLOW)

            if os.uname().machine == 'x86_64':
                # Allow Python and try-syscall to be different word sizes
                denylist.add_arch(seccomp.Arch.X86)

            # Using ECONNREFUSED here because it's unlikely that any of
            # these syscalls will legitimately fail with that code, so
            # if they fail like this, it will be as a result of seccomp.
            denylist.add_rule(seccomp.ERRNO(errno.ECONNREFUSED), 'chmod')
            denylist.add_rule(seccomp.ERRNO(errno.ECONNREFUSED), 'chroot')
            denylist.add_rule(seccomp.ERRNO(errno.ECONNREFUSED), 'prctl')
            denylist.add_rule(
                seccomp.ERRNO(errno.ECONNREFUSED), 'ioctl',
                seccomp.Arg(1, seccomp.MASKED_EQ, 0xffffffff, termios.TIOCSTI),
            )

            denylist.export_bpf(denylist_temp)

            for syscall in TRY_SYSCALLS:
                print('# denylist vs. {}'.format(syscall))
                denylist_temp.seek(0, os.SEEK_SET)

                completed = subprocess.run(
                    [
                        self.bwrap,
                        '--ro-bind', '/', '/',
                        '--seccomp', str(denylist_temp.fileno()),
                        self.try_syscall, syscall,
                    ],
                    pass_fds=(denylist_temp.fileno(),),
                    stdin=subprocess.DEVNULL,
                    stdout=subprocess.DEVNULL,
                    stderr=2,
                )

                if (
                    syscall == 'ioctl TIOCSTI CVE-2019-10063'
                    and completed.returncode == errno.ENOENT
                ):
                    print('# Cannot test 64-bit syscall parameter on 32-bit')
                    continue

                if syscall == 'clone3':
                    # If the kernel supports it, we didn't block it so
                    # it fails with EFAULT. If the kernel doesn't support it,
                    # it'll fail with ENOSYS instead.
                    self.assertIn(
                        completed.returncode,
                        (errno.ENOSYS, errno.EFAULT),
                    )
                elif syscall in ('ioctl TIOCNOTTY', 'listen'):
                    # Not on the denylist
                    self.assertEqual(completed.returncode, errno.EBADF)
                else:
                    # We blocked all of these
                    self.assertEqual(completed.returncode, errno.ECONNREFUSED)

    def test_seccomp_stacked(self, allowlist_first=False) -> None:
        with tempfile.TemporaryFile(
        ) as allowlist_temp, tempfile.TemporaryFile(
        ) as denylist_temp:
            # This filter is a simplified version of what Flatpak wants

            allowlist = seccomp.SyscallFilter(seccomp.ERRNO(errno.ENOSYS))
            denylist = seccomp.SyscallFilter(seccomp.ALLOW)

            if os.uname().machine == 'x86_64':
                # Allow Python and try-syscall to be different word sizes
                allowlist.add_arch(seccomp.Arch.X86)
                denylist.add_arch(seccomp.Arch.X86)

            for syscall in ALLOWED:
                try:
                    allowlist.add_rule(seccomp.ALLOW, syscall)
                except Exception as e:
                    print('# Cannot add {} to allowlist: {!r}'.format(syscall, e))

            denylist.add_rule(seccomp.ERRNO(errno.ECONNREFUSED), 'chmod')
            denylist.add_rule(seccomp.ERRNO(errno.ECONNREFUSED), 'chroot')
            denylist.add_rule(
                seccomp.ERRNO(errno.ECONNREFUSED), 'ioctl',
                seccomp.Arg(1, seccomp.MASKED_EQ, 0xffffffff, termios.TIOCSTI),
            )

            # All seccomp programs except the last must allow prctl(),
            # because otherwise we wouldn't be able to add the remaining
            # seccomp programs. We document that the last program can
            # block prctl, so test that.
            if allowlist_first:
                denylist.add_rule(seccomp.ERRNO(errno.ECONNREFUSED), 'prctl')

            allowlist.export_bpf(allowlist_temp)
            denylist.export_bpf(denylist_temp)

            for syscall in TRY_SYSCALLS:
                print('# stacked vs. {}'.format(syscall))
                allowlist_temp.seek(0, os.SEEK_SET)
                denylist_temp.seek(0, os.SEEK_SET)

                if allowlist_first:
                    fds = [allowlist_temp.fileno(), denylist_temp.fileno()]
                else:
                    fds = [denylist_temp.fileno(), allowlist_temp.fileno()]

                completed = subprocess.run(
                    [
                        self.bwrap,
                        '--ro-bind', '/', '/',
                        '--add-seccomp-fd', str(fds[0]),
                        '--add-seccomp-fd', str(fds[1]),
                        self.try_syscall, syscall,
                    ],
                    pass_fds=fds,
                    stdin=subprocess.DEVNULL,
                    stdout=subprocess.DEVNULL,
                    stderr=2,
                )

                if (
                    syscall == 'ioctl TIOCSTI CVE-2019-10063'
                    and completed.returncode == errno.ENOENT
                ):
                    print('# Cannot test 64-bit syscall parameter on 32-bit')
                    continue

                if syscall == 'ioctl TIOCNOTTY':
                    # Not denied by the denylist, and allowed by the allowlist
                    self.assertEqual(completed.returncode, errno.EBADF)
                elif syscall in ('clone3', 'listen'):
                    # We didn't deny these, so the denylist has no effect
                    # and we fall back to the allowlist, which doesn't
                    # include them either.
                    # clone3 might also be failing with ENOSYS because
                    # the kernel genuinely doesn't support it.
                    self.assertEqual(completed.returncode, errno.ENOSYS)
                elif syscall == 'chroot':
                    # This is denied by the denylist *and* not allowed by
                    # the allowlist. The result depends which one we added
                    # first: the most-recently-added filter "wins".
                    if allowlist_first:
                        self.assertEqual(
                            completed.returncode,
                            errno.ECONNREFUSED,
                        )
                    else:
                        self.assertEqual(completed.returncode, errno.ENOSYS)
                elif syscall == 'prctl':
                    # We can only put this on the denylist if the denylist
                    # is the last to be added.
                    if allowlist_first:
                        self.assertEqual(
                            completed.returncode,
                            errno.ECONNREFUSED,
                        )
                    else:
                        self.assertEqual(completed.returncode, errno.EFAULT)
                else:
                    # chmod is allowed by the allowlist but blocked by the
                    # denylist. Denying takes precedence over allowing,
                    # regardless of order.
                    self.assertEqual(completed.returncode, errno.ECONNREFUSED)

    def test_seccomp_stacked_allowlist_first(self) -> None:
        self.test_seccomp_stacked(allowlist_first=True)

    def test_seccomp_invalid(self) -> None:
        with tempfile.TemporaryFile(
        ) as allowlist_temp, tempfile.TemporaryFile(
        ) as denylist_temp:
            completed = subprocess.run(
                [
                    self.bwrap,
                    '--ro-bind', '/', '/',
                    '--add-seccomp-fd', '-1',
                    'true',
                ],
                stdin=subprocess.DEVNULL,
                stdout=subprocess.DEVNULL,
                stderr=subprocess.PIPE,
            )
            self.assertIn(b'bwrap: Invalid fd: -1\n', completed.stderr)
            self.assertEqual(completed.returncode, 1)

            completed = subprocess.run(
                [
                    self.bwrap,
                    '--ro-bind', '/', '/',
                    '--seccomp', '0a',
                    'true',
                ],
                stdin=subprocess.DEVNULL,
                stdout=subprocess.DEVNULL,
                stderr=subprocess.PIPE,
            )
            self.assertIn(b'bwrap: Invalid fd: 0a\n', completed.stderr)
            self.assertEqual(completed.returncode, 1)

            completed = subprocess.run(
                [
                    self.bwrap,
                    '--ro-bind', '/', '/',
                    '--add-seccomp-fd', str(denylist_temp.fileno()),
                    '--seccomp', str(allowlist_temp.fileno()),
                    'true',
                ],
                pass_fds=(allowlist_temp.fileno(), denylist_temp.fileno()),
                stdin=subprocess.DEVNULL,
                stdout=subprocess.DEVNULL,
                stderr=subprocess.PIPE,
            )
            self.assertIn(
                b'bwrap: --seccomp cannot be combined with --add-seccomp-fd\n',
                completed.stderr,
            )
            self.assertEqual(completed.returncode, 1)

            completed = subprocess.run(
                [
                    self.bwrap,
                    '--ro-bind', '/', '/',
                    '--seccomp', str(allowlist_temp.fileno()),
                    '--add-seccomp-fd', str(denylist_temp.fileno()),
                    'true',
                ],
                pass_fds=(allowlist_temp.fileno(), denylist_temp.fileno()),
                stdin=subprocess.DEVNULL,
                stdout=subprocess.DEVNULL,
                stderr=subprocess.PIPE,
            )
            self.assertIn(
                b'--add-seccomp-fd cannot be combined with --seccomp',
                completed.stderr,
            )
            self.assertEqual(completed.returncode, 1)

            completed = subprocess.run(
                [
                    self.bwrap,
                    '--ro-bind', '/', '/',
                    '--add-seccomp-fd', str(allowlist_temp.fileno()),
                    '--add-seccomp-fd', str(allowlist_temp.fileno()),
                    'true',
                ],
                pass_fds=(allowlist_temp.fileno(), allowlist_temp.fileno()),
                stdin=subprocess.DEVNULL,
                stdout=subprocess.DEVNULL,
                stderr=subprocess.PIPE,
            )
            self.assertIn(
                b"bwrap: Can't read seccomp data: ",
                completed.stderr,
            )
            self.assertEqual(completed.returncode, 1)

            allowlist_temp.write(b'\x01')
            allowlist_temp.seek(0, os.SEEK_SET)
            completed = subprocess.run(
                [
                    self.bwrap,
                    '--ro-bind', '/', '/',
                    '--add-seccomp-fd', str(denylist_temp.fileno()),
                    '--add-seccomp-fd', str(allowlist_temp.fileno()),
                    'true',
                ],
                pass_fds=(allowlist_temp.fileno(), denylist_temp.fileno()),
                stdin=subprocess.DEVNULL,
                stdout=subprocess.DEVNULL,
                stderr=subprocess.PIPE,
            )
            self.assertIn(
                b'bwrap: Invalid seccomp data, must be multiple of 8\n',
                completed.stderr,
            )
            self.assertEqual(completed.returncode, 1)


def main():
    logging.basicConfig(level=logging.DEBUG)

    try:
        from tap.runner import TAPTestRunner
    except ImportError:
        TAPTestRunner = None    # type: ignore

    if TAPTestRunner is not None:
        runner = TAPTestRunner()
        runner.set_stream(True)
        unittest.main(testRunner=runner)
    else:
        print('# tap.runner not available, using simple TAP output')
        print('1..1')
        program = unittest.main(exit=False)
        if program.result.wasSuccessful():
            print('ok 1 - %r' % program.result)
        else:
            print('not ok 1 - %r' % program.result)
            sys.exit(1)

if __name__ == '__main__':
    main()
